/*
    CharityWorld: I1.c
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/*

I1.c

Run through all base 10 numbers from start.0...0 to start.9...9 where
the number of digits to replace 0...0 by is given by digit10. Use each
of these numbers to provide a value for G, the generosity parameter,
to see if CharityWorld would terminate and converge in a 3 x 3 grid
with the winner at (2, 2), jackpot equal to 9G, initial wealth 0 and
SSA agent class. Check for floating point errors in:

1. Conversion of G from decimal to binary
2. Conversion of J from decimal to binary
3. Calculation of ticket price
4. Adding wealth to winner
5. Subtracting wealth from winner
6. Adding wealth to losers
7. Subtracting wealth from losers

Also check that:

8. Simulation terminates after 1 cycle.
9. Simulation converges after 1 cycle.
10. Ticket price equal to G.
11. Winner has zero wealth at the end.
12. Losers have zero wealth at the end.

*/

#include <stdio.h>		/* fprintf, stderr, sprintf */
#include <stdlib.h>		/* atoi, malloc, abort, atof */
#include <ieeefp.h>		/* fpsetsticky, fpgetsticky */
#include <float.h>		/* DBL_DIG */
#include <errno.h>		/* perror */
#include <string.h>		/* strlen */

#define NCHKERR 12		/* Number of errors and checks for each G */
#define NERR 7			/* Number of errors for each G */

typedef union {
  double d;
  float f;
} FP;

typedef struct match_list {
  char match[NCHKERR + 1];
  struct match_list *next;
} MLST;

struct options {
  int single;
  int quiet;
  MLST *filter;
} opts;

int process_options(int argc, char **argv, void (* usg)(char *));
void default_options(void);
void print_options(void);
int options_match(MLST *rslt);
void usage(char *cmd);

int main(int argc, char **argv) {
  int digit10;			/* Number of base 10 digits in the fraction */
  int i;
  int *counter;			/* Array to count through the base 10 digits */
  int j;	
  int start;			/* Integer part of the total fraction */
  int bufsz;			/* Size of buffer needed to sprintf number */
  int bufstart;			/* Place in buffer to start fraction */
  FP G;				/* Generosity model parameter */
  char *buf;			/* Buffer for number */
  char dbl2str[DBL_DIG + 11];	/* Space for a double in a string */
  char temp[4];
  int argn;			/* Offset from argv[0] for first non-option */
  unsigned long long nmatch = 0LLU;
				/* Count of number of matching lines */

  argn = process_options(argc, argv, usage);

  if(argc - argn < 3) {		/* Check all arguments have been provided */
    usage(argv[0]);
  }

  /* Check options are valid */

  if(opts.quiet && opts.filter == NULL) {
    printf("Summary output and no filters: All numbers match!\n");
    return 0;
  }

  /* Initialise variables */

  digit10 = atoi(argv[argn + 2]);
  start = atoi(argv[argn + 1]);

  counter = (int *)malloc(digit10 * sizeof(int));
  if(counter == NULL) {
    perror("Memory allocation of counter");
    abort();
  }

  for(i = 0; i < digit10; i++) {
    counter[i] = 0;
  }

  bufsz = snprintf(temp, 3, "%d.%0*d", start, digit10, 0);
  buf = (char *)malloc((bufsz + 1) * sizeof(char));
  if(buf == NULL) {
    perror("Memory allocation of buf");
    abort();
  }

  bufstart = sprintf(buf, "%d.", start);

  /* Print a header line or two */

  print_options();

  if(!opts.quiet) {
    printf("Errors:\n"
	   "\tG -- decimal to binary conversion of generosity parameter FPE\n"
	   "\tJ -- decimal to binary conversion of jackpot parameter FPE\n"
	   "\tP -- FPE during calculation of ticket price\n"
	   "\tW -- FPE adding winnings to winner\n"
	   "\tw -- FPE subtracting generosity from winner\n"
	   "\tl -- FPE subtracting ticket price from losers\n"
	   "\tL -- FPE adding ticket price to losers\n"
	   "Checks:\n"
	   "\tT -- termination: no loser or winner awareness condition met\n"
	   "\tC -- convergence: winner and loser have same wealth at the end\n"
	   "\tE -- ticket price equal to generosity\n"
	   "\tZ -- winner has zero wealth at the end\n"
	   "\tz -- losers have zero wealth at the end\n\n");
    printf("%*s | Errors        | Checks\n%*s | G J P W w l L | T C E Z z\n",
	   bufsz, "G", bufsz, " ");
    for(i = 0; i < bufsz; i++) {
      printf("-");
    }
    printf("-+---------------+----------\n");
  }

  /* Loop through the fractions from 0...0 to 9...9 */
    
  while(1) {
    FP ticket_price;
    FP winners_wealth;
    FP losers_wealth;
    FP J;
    int k;
    fp_except Gerr;
    fp_except Jerr;
    fp_except tperr;
    fp_except waerr, wserr;
    fp_except laerr, lserr;
    int terminate, converge, equal, winner_zero, losers_zero;
    MLST result;

    /* Build up the G parameter in a string as start.fraction -- we
       only need to print the fraction part -- the start bit was
       printed to buf before the loop began. */

    for(j = 0; j < digit10; j++) {
      sprintf(&buf[j + bufstart], "%d", counter[j] % 10);
    }

    /* Convert G from decimal to binary, and check for errors. */
    
    fpsetsticky(0);
    if(opts.single) G.f = (float)atof(buf);
    else G.d = atof(buf);
    Gerr = fpgetsticky();

    /* Now calculate 9G, and put it in a string, as though it was in a
       parameter file, ready for conversion into J. */

    if(opts.single) {
      sprintf(dbl2str, "%.*e", FLT_DIG - 1, (double)(9.0F * G.f));
    }
    else sprintf(dbl2str, "%.*e", DBL_DIG - 1,  9.0 * G.d);

    /* Convert J from decimal to binary, and check for errors. */

    fpsetsticky(0);
    if(opts.single) J.f = (float)atof(dbl2str);
    else J.d = atof(dbl2str);
    Jerr = fpgetsticky();

    /* Calculate the ticket price, and check for errors. */

    fpsetsticky(0);
    if(opts.single) ticket_price.f = J.f / 9.0F;
    else ticket_price.d = J.d / 9.0;
    tperr = fpgetsticky();

    /* Add and subtract the winner's wealth and check for errors. We
       assume that the winner will always donate to the last loser,
       regardless of errors in floating point arithmetic. This is not
       unreasonable. At this point, the last loser has wealth
       -ticket_price, whilst each loser has wealth G - ticket_price,
       and the winner has wealth (8 * ticket_price) - (7 * G). The
       total wealth is therefore zero (or very close to it), and 9 *
       winners_wealth is definitely positive. The winner will
       therefore be supply-aware, and will approach the last loser. */

    if(opts.single) winners_wealth.f = 0.0F;
    else winners_wealth.d = 0.0;
    fpsetsticky(0);
    for(k = 0; k < 8; k++) {
      if(opts.single) winners_wealth.f += ticket_price.f;
      else winners_wealth.d += ticket_price.d;
    }
    waerr = fpgetsticky();
    fpsetsticky(0);
    for(k = 0; k < 8; k++) {
      if(opts.single) winners_wealth.f -= G.f;
      else winners_wealth.d -= G.d;
    }
    wserr = fpgetsticky();

    /* Add and subtract the losers' wealth and check for errors. */

    if(opts.single) losers_wealth.f = 0.0F;
    else losers_wealth.d = 0.0;
    fpsetsticky(0);
    if(opts.single) losers_wealth.f -= ticket_price.f;
    else losers_wealth.d -= ticket_price.d;
    lserr = fpgetsticky();
    fpsetsticky(0);
    if(opts.single) losers_wealth.f += G.f;
    else losers_wealth.d += G.d;
    laerr = fpgetsticky();

    /* Check for termination: Does the winner still want to donate? */

    terminate = 1;
    if(opts.single) {
      if(winners_wealth.f * 9.0F > 
	 winners_wealth.f + losers_wealth.f + losers_wealth.f + losers_wealth.f
	 + losers_wealth.f + losers_wealth.f + losers_wealth.f
	 + losers_wealth.f + losers_wealth.f) {
	terminate = 0;
      }
    }
    else {
      if(winners_wealth.d * 9.0 > 
	 winners_wealth.d + losers_wealth.d + losers_wealth.d + losers_wealth.d
	 + losers_wealth.d + losers_wealth.d + losers_wealth.d
	 + losers_wealth.d + losers_wealth.d) {
	terminate = 0;
      }
    }

    /* Check for termination cont'd: Does a loser what to donate? */

    if(opts.single) {
      if(losers_wealth.f * 9.0F > 
	 winners_wealth.f + losers_wealth.f + losers_wealth.f + losers_wealth.f
	 + losers_wealth.f + losers_wealth.f + losers_wealth.f
	 + losers_wealth.f + losers_wealth.f) {
	terminate = 0;
      }
    }
    else {
      if(losers_wealth.d * 9.0 > 
	 winners_wealth.d + losers_wealth.d + losers_wealth.d + losers_wealth.d
	 + losers_wealth.d + losers_wealth.d + losers_wealth.d
	 + losers_wealth.d + losers_wealth.d) {
	terminate = 0;
      }
    }

    /* Check for convergence: winner's wealth == losers' wealth? */

    converge = 0;
    if(opts.single) {
      if(winners_wealth.f == losers_wealth.f) converge = 1;
    }
    else {
      if(winners_wealth.d == losers_wealth.d) converge = 1;
    }

    /* Check ticket price == G */

    equal = 0;
    if(opts.single) {
      if(ticket_price.f == G.f) equal = 1;
    }
    else {
      if(ticket_price.d == G.d) equal = 1;
    }

    /* Check winner's wealth == 0 */

    winner_zero = 0;
    if(opts.single) {
      if(winners_wealth.f == 0.0F) winner_zero = 1;
    }
    else {
      if(winners_wealth.d == 0.0) winner_zero = 1;
    }

    /* Check losers' wealth == 0 */
   
    losers_zero = 0;
    if(opts.single) {
      if(losers_wealth.f == 0.0F) losers_zero = 1;
    }
    else {
      if(losers_wealth.d == 0.0) losers_zero = 1;
    }

    /* Print out a summary of the run, including all errors */

    result.match[0] = (Gerr == 0) ? 'n' : 'y';
    result.match[1] = (Jerr == 0) ? 'n' : 'y';
    result.match[2] = (tperr == 0) ? 'n' : 'y';
    result.match[3] = (waerr == 0) ? 'n' : 'y';
    result.match[4] = (wserr == 0) ? 'n' : 'y';
    result.match[5] = (lserr == 0) ? 'n' : 'y';
    result.match[6] = (laerr == 0) ? 'n' : 'y';

    result.match[7] = terminate ? 'y' : 'n';
    result.match[8] = converge ? 'y' : 'n';
    result.match[9] = equal ? 'y' : 'n';
    result.match[10] = winner_zero ? 'y' : 'n';
    result.match[11] = losers_zero ? 'y' : 'n';

    result.match[NCHKERR] = '\0';

    if(options_match(&result)) {
      if(!opts.quiet) {
	printf("%s | %c %c %c %c %c %c %c | %c %c %c %c %c\n", buf,
	       (Gerr == 0) ? 'n' : 'Y',
	       (Jerr == 0) ? 'n' : 'Y',
	       (tperr == 0) ? 'n' : 'Y',
	       (waerr == 0) ? 'n' : 'Y',
	       (wserr == 0) ? 'n' : 'Y',
	       (lserr == 0) ? 'n' : 'Y',
	       (laerr == 0) ? 'n' : 'Y',
	       terminate ? 'y' : 'N',
	       converge ? 'y' : 'N',
	       equal ? 'y' : 'N',
	       winner_zero ? 'y' : 'N',
	       losers_zero ? 'y' : 'N');
      }
      nmatch++;
    }

    /* Iterate the loop */

    for(j = digit10 - 1; j >= 0; j--) {
      counter[j]++;
      if(j > 0 && counter[j] > 9) {
	counter[j] = 0;
	continue;
      }
      break;
    }

    if(counter[0] > 9) break;
  }

  /* Print a summary */

  if(!opts.quiet) {
    for(i = 0; i < bufsz; i++) {
      printf("-");
    }
    printf("-+---------------+----------\n\n");
  }
  printf("%llu of 10^%d numbers matched\n", nmatch, digit10);

  /* Free allocated memory */
 
  free(buf);
  free(counter);
  while(opts.filter != NULL) {
    MLST *tmp = opts.filter->next;
    free(opts.filter);
    opts.filter = tmp;
  }

  return 0;
}

/*

usage:

function to display a usage for this program

*/

void usage(char *cmd) {
  fprintf(stderr, "%s [options] <start> <base 10 digits>\n", cmd);
  fprintf(stderr, "Options are:\n"
	  "\t-C\tAdd a filter to print lines that terminate and converge\n"
	  "\t\twhen the ticket price is not equal to the generosity\n"
	  "\t-c\tAdd a filter to print lines that terminate and converge\n"
	  "\t-e\tAdd filters to print lines with a floating point error\n"
	  "\t-i\tAdd filters to print lines that converge without\n"
	  "\t\tterminating or vice versa\n"
	  "\t-m f\tOnly print lines matching f, which is a series of\n"
	  "\t\t%d characters containing y, n or . (wildcard). Match\n"
	  "\t\tis not case sensitive and you can use many -m options\n"
	  "\t\tfor a disjunction of filters.\n"
	  "\t-q\tQuiet output -- no table of results, just a summary\n"
	  "\t-s\tUse single instead of double precision\n"
	  "\t-T\tAdd a filter to print lines that terminate and converge\n"
	  "\t\twith no floating point errors\n"
	  "\t-t\tAdd filters to print lines that terminate and converge\n"
	  "\t\twith floating point errors\n", NCHKERR);
  exit(1);
}

/*

options_match:

Check the result of a particular value for G against any matches
specified by the user with -m options to see if the results should be
printed.

*/

int options_match(MLST *rslt) {
  MLST *iterator;

  if(opts.filter == NULL) return 1;
				/* Deal with case where no -m option given */
  for(iterator = opts.filter; iterator != NULL; iterator = iterator->next) {
    int i;
    int ok = 1;

    for(i = 0; i < NCHKERR; i++) {
      if(iterator->match[i] == '.') continue;
      if(iterator->match[i] != rslt->match[i]) {
	ok = 0;
	break;
      }
    }
    if(ok) return 1;
  }
  return 0;
}

/*

default_options:

set all the options to their default values

*/

void default_options(void) {
  opts.single = 0;
  opts.quiet = 0;
  opts.filter = NULL;
}

/*

print_options:

print all option settings

*/

void print_options(void) {
  printf("Options:\n");
  printf("\tfilters:");
  if(opts.filter == NULL) {
    printf(" [NONE]\n");
  }
  else {
    MLST *fltr = opts.filter;
    printf("\n\t\tGJPWwlLTCEZz\n");
    while(fltr != NULL) {
      printf("\t\t%s\n", fltr->match);
      fltr = fltr->next;
    }
  }
  printf("\tfull/summary output: [%s]\n",
	 opts.quiet ? "SUMMARY" : "FULL");
  printf("\tsingle/double precision: [%s]\n",
	 opts.single ? "SINGLE" : "DOUBLE");

  printf("\n");
}

/*

new_filter:

Allocate a new filter making opts.filter point to the newly allocated one.

*/

void new_filter(char *flt) {
  MLST *tmp;
  int i;

  tmp = opts.filter;
  opts.filter = (MLST *)malloc(sizeof(MLST));
  if(opts.filter == NULL) {
    perror("Allocating memory for filter");
    abort();
  }
  opts.filter->next = tmp;
  for(i = 0; i < NCHKERR; i++) {
    if(flt[i] == 'Y' || flt[i] == 'y') {
      opts.filter->match[i] = 'y';
    }
    else if(flt[i] == 'N' || flt[i] == 'n') {
      opts.filter->match[i] = 'n';
    }
    else if(flt[i] == '.') {
      opts.filter->match[i] = '.';
    }
    else {
      fprintf(stderr, "Error in match string %s, character %d: invalid "
	      "match token %c\n", flt, i + 1, flt[i]);
      abort();
    }
  }
  opts.filter->match[NCHKERR] = '\0';
}

/*

process_options:

function to process the options passed in as arguments. Uses the opts
global variable to store all the options.

*/

int process_options(int argc, char **argv, void (* usg)(char *)) {
  int argn = 0;
  int i;
  char *fpefilters[NERR] = {
    "y...........",
    ".y..........",
    "..y.........",
    "...y........",
    "....y.......",
    ".....y......",
    "......y....."
  };
  char *termfilter = ".......yy...";
  char *fpetermfilters[NERR] = {
    "y......yy...",
    ".y.....yy...",
    "..y....yy...",
    "...y...yy...",
    "....y..yy...",
    ".....y.yy...",
    "......yyy..."
  };
  char *nofpetermfilter = "nnnnnnnyy...";
  
  default_options();
  while(argn + 1 < argc && argv[argn + 1][0] == '-') {
    argn++;
    switch(argv[argn][1]) {
    case 'c':
      new_filter(termfilter);
      break;
    case 'C':
      new_filter(".......yyn..");
      break;
    case 'e':
      for(i = 0; i < NERR; i++) {
	new_filter(fpefilters[i]);
      }
      break;
    case 'i':
      new_filter(".......yn...");
      new_filter(".......ny...");
      break;
    case 'm':
      if(argn + 1 >= argc) {
	fprintf(stderr, "Option -m needs a filter as an argument\n");
	usg(argv[0]);
      }
      argn++;
      new_filter(argv[argn]);
      break;
    case 'q':
      opts.quiet = 1;
      break;
    case 's':
      opts.single = 1;
      break;
    case 't':
      for(i = 0; i < NERR; i++) {
	new_filter(fpetermfilters[i]);
      }
      break;
    case 'T':
      new_filter(nofpetermfilter);
      break;
    default:
      fprintf(stderr, "Option %s not recognised\n", argv[argn]);
      usg(argv[0]);
    }
  }
  return argn;
}
